import kotlinx.coroutines.*
import kotlin.math.PI
import kotlin.math.sin
import kotlin.time.*
import kotlin.time.TimeSource.*

const val MAXSTEPS = 10_000_000


@OptIn(ExperimentalTime::class)
fun main() {
    // Integral über x + sin(x) für den Wertebereich von 0 bis 2 * Pi
    // das exakte Ergebnis lautet (2 * Pi)^2 / 2, also  19.7392...
    val result = measureAverage {
        integrateNumeric(start = 0.0, end = 2 * PI) { x -> x + sin(x) }
    }
    println("Ergebnis:   ${result.value}")
    println("Rechenzeit: ${result.duration}")

    val result2 = measureAverage {
        integrateParallel(start = 0.0, end = 2 * PI) { x -> x + sin(x) }
    }
    println("Ergebnis:   ${result2.value}")
    println("Rechenzeit: ${result2.duration}")
}

// Funktion 10x aufrufen, Durchschnittswert der Laufzeit berechnen
@OptIn(ExperimentalTime::class)
fun <T> measureAverage(block: () -> T ) : TimedValue<T> {
    val passes = 10
    val now = Monotonic.markNow()
    val result = block()   // 1. Aufruf
    for (i in 1 until passes)    // noch 9 Aufrufe
        block()
    val dur = now.elapsedNow() / passes
    return TimedValue(result, dur)
}

// numerisches Integral über die Funktion 'func' um
// Wertebereich von 'start' bis 'end'
fun integrateNumeric(start: Double, end: Double, func:  (Double) -> Double): Double {
    val steps = MAXSTEPS  // Anzahl der Schritte für die numerische Integration
    var result = 0.0
    if (start == end) return 0.0
    val delta = (end - start) / (steps - 1)
    for (i in 0 until steps) {
        val x = start + delta * i
        result += func(x)
    }
    return result / steps * (end - start)
}


fun integrateParallel(start: Double, end: Double, func: (Double) -> Double): Double {
    if (start == end) return 0.0
    val steps = MAXSTEPS
    val segments = 100     // Anzahl der parallelen Koroutinen
    val results  = mutableListOf<Deferred<Double>>()
    var sum = 0.0

    // Variante 1
    runBlocking {
        // 100 Segmente in parallelen Koroutinen berechnen
        for (segment in 0 until segments)
            results += async(Dispatchers.Default) {
                sumSegment(start, end, steps, segments, segment, func)
            }
        // Ergebnisse abwarten und summieren
        sum = results.awaitAll().sum()
    }

    /*
    // Variante 2: mit einem eigenen CoroutineScope
    val myscope = CoroutineScope(Dispatchers.Default)
    runBlocking(myscope.coroutineContext) {
        for (segment in 0 until segments) {
            results += myscope.async {
                sumSegment(start, end, steps, segments, segment, func)
            }
        }
        sum = results.awaitAll().sum()
    }
    */

    return sum / steps * (end - start)
}


fun sumSegment(start: Double,
               end: Double,
               steps: Int,
               segments: Int,
               thisSegment: Int,
               func: (Double) -> Double): Double
{
    val stepsInSegment = steps / segments
    val delta = (end - start) / (steps - 1)
    var result = 0.0
    for (i in 0 until stepsInSegment) {
        val x = start + delta * (i + thisSegment * stepsInSegment)
        result += func(x)
    }
    return result
}

